@@ -1,11 +1,11 @@
+# frozen_string_literal: true
+
 require 'test/unit'
 require 'pathname'
 
 require 'fileutils'
 require 'tmpdir'
-require 'enumerator'
 
-require_relative '../ruby/envutil'
 
 class TestPathname < Test::Unit::TestCase
   def self.define_assertion(name, linenum, &block)
@@ -88,6 +88,10 @@ class TestPathname < Test::Unit::TestCase
     defassert(:cleanpath_aggressive, '/',       '///a/../..')
   end
 
+  if DOSISH
+    defassert(:cleanpath_aggressive, 'c:/foo/bar', 'c:\\foo\\bar')
+  end
+
   def cleanpath_conservative(path)
     Pathname.new(path).cleanpath(true).to_s
   end
@@ -124,6 +128,10 @@ class TestPathname < Test::Unit::TestCase
   defassert(:cleanpath_conservative, '/a',     '/../.././../a')
   defassert(:cleanpath_conservative, 'a/b/../../../../c/../d', 'a/b/../../../../c/../d')
 
+  if DOSISH
+    defassert(:cleanpath_conservative, 'c:/foo/bar', 'c:\\foo\\bar')
+  end
+
   if DOSISH_UNC
     defassert(:cleanpath_conservative, '//',     '//')
   else
@@ -183,8 +191,8 @@ class TestPathname < Test::Unit::TestCase
 
   if DOSISH
     defassert(:del_trailing_separator, "a", "a\\")
-    defassert(:del_trailing_separator, "\225\\".force_encoding("cp932"), "\225\\\\".force_encoding("cp932"))
-    defassert(:del_trailing_separator, "\225".force_encoding("cp437"), "\225\\\\".force_encoding("cp437"))
+    defassert(:del_trailing_separator, "\225\\".dup.force_encoding("cp932"), "\225\\\\".dup.force_encoding("cp932"))
+    defassert(:del_trailing_separator, "\225".dup.force_encoding("cp437"), "\225\\\\".dup.force_encoding("cp437"))
   end
 
   def test_plus
@@ -213,6 +221,10 @@ class TestPathname < Test::Unit::TestCase
 
   defassert(:plus, 'a//b/d//e', 'a//b/c', '../d//e')
 
+  def test_slash
+    assert_kind_of(Pathname, Pathname("a") / Pathname("b"))
+  end
+
   def test_parent
     assert_equal(Pathname("."), Pathname("a").parent)
   end
@@ -234,6 +246,16 @@ class TestPathname < Test::Unit::TestCase
   def test_join
     r = Pathname("a").join(Pathname("b"), Pathname("c"))
     assert_equal(Pathname("a/b/c"), r)
+    r = Pathname("/a").join(Pathname("b"), Pathname("c"))
+    assert_equal(Pathname("/a/b/c"), r)
+    r = Pathname("/a").join(Pathname("/b"), Pathname("c"))
+    assert_equal(Pathname("/b/c"), r)
+    r = Pathname("/a").join(Pathname("/b"), Pathname("/c"))
+    assert_equal(Pathname("/c"), r)
+    r = Pathname("/a").join("/b", "/c")
+    assert_equal(Pathname("/c"), r)
+    r = Pathname("/foo/var").join()
+    assert_equal(Pathname("/foo/var"), r)
   end
 
   def test_absolute
@@ -319,10 +341,10 @@ class TestPathname < Test::Unit::TestCase
 
   def has_symlink?
     begin
-      File.symlink(nil, nil)
-    rescue NotImplementedError
+      File.symlink("", "")
+    rescue NotImplementedError, Errno::EACCES
       return false
-    rescue TypeError
+    rescue Errno::ENOENT
     end
     return true
   end
@@ -381,6 +403,7 @@ class TestPathname < Test::Unit::TestCase
       File.symlink("f/g", "h")
       assert_equal("#{dir}/f/g", realpath("h"))
       File.chmod(0000, "f")
+      next if File.readable?("f")
       assert_raise(Errno::EACCES) { realpath("h") }
       File.chmod(0755, "f")
     }
@@ -409,7 +432,7 @@ class TestPathname < Test::Unit::TestCase
   end
 
   def descend(path)
-    Pathname.new(path).enum_for(:descend).map {|v| v.to_s }
+    Pathname.new(path).descend.map(&:to_s)
   end
 
   defassert(:descend, %w[/ /a /a/b /a/b/c], "/a/b/c")
@@ -418,7 +441,7 @@ class TestPathname < Test::Unit::TestCase
   defassert(:descend, %w[a/], "a/")
 
   def ascend(path)
-    Pathname.new(path).enum_for(:ascend).map {|v| v.to_s }
+    Pathname.new(path).ascend.map(&:to_s)
   end
 
   defassert(:ascend, %w[/a/b/c /a/b /a /], "/a/b/c")
@@ -426,6 +449,14 @@ class TestPathname < Test::Unit::TestCase
   defassert(:ascend, %w[./a/b/c ./a/b ./a .], "./a/b/c")
   defassert(:ascend, %w[a/], "a/")
 
+  def test_blockless_ascend_is_enumerator
+    assert_kind_of(Enumerator, Pathname.new('a').ascend)
+  end
+
+  def test_blockless_descend_is_enumerator
+    assert_kind_of(Enumerator, Pathname.new('a').descend)
+  end
+
   def test_initialize
     p1 = Pathname.new('a')
     assert_equal('a', p1.to_s)
@@ -537,6 +568,11 @@ class TestPathname < Test::Unit::TestCase
     assert_include([true, false], r)
   end
 
+  def test_mountpoint_enoent
+    r = Pathname("/nonexistent").mountpoint?
+    assert_equal false, r
+  end
+
   def test_destructive_update
     path = Pathname.new("a")
     path.to_s.replace "b"
@@ -551,16 +587,16 @@ class TestPathname < Test::Unit::TestCase
     obj = Pathname.new("a"); assert_same(obj, obj.taint)
     obj = Pathname.new("a"); assert_same(obj, obj.untaint)
 
-    assert_equal(false, Pathname.new("a"      )           .tainted?)
-    assert_equal(false, Pathname.new("a"      )      .to_s.tainted?)
-    assert_equal(true,  Pathname.new("a"      ).taint     .tainted?)
-    assert_equal(true,  Pathname.new("a"      ).taint.to_s.tainted?)
-    assert_equal(true,  Pathname.new("a".taint)           .tainted?)
-    assert_equal(true,  Pathname.new("a".taint)      .to_s.tainted?)
-    assert_equal(true,  Pathname.new("a".taint).taint     .tainted?)
-    assert_equal(true,  Pathname.new("a".taint).taint.to_s.tainted?)
+    assert_equal(false, Pathname.new("a"          )           .tainted?)
+    assert_equal(false, Pathname.new("a"          )      .to_s.tainted?)
+    assert_equal(true,  Pathname.new("a"          ).taint     .tainted?)
+    assert_equal(true,  Pathname.new("a"          ).taint.to_s.tainted?)
+    assert_equal(true,  Pathname.new("a".dup.taint)           .tainted?)
+    assert_equal(true,  Pathname.new("a".dup.taint)      .to_s.tainted?)
+    assert_equal(true,  Pathname.new("a".dup.taint).taint     .tainted?)
+    assert_equal(true,  Pathname.new("a".dup.taint).taint.to_s.tainted?)
 
-    str = "a"
+    str = "a".dup
     path = Pathname.new(str)
     str.taint
     assert_equal(false, path     .tainted?)
@@ -573,7 +609,7 @@ class TestPathname < Test::Unit::TestCase
     assert_equal(false, Pathname.new("a").taint.untaint     .tainted?)
     assert_equal(false, Pathname.new("a").taint.untaint.to_s.tainted?)
 
-    str = "a".taint
+    str = "a".dup.taint
     path = Pathname.new(str)
     str.untaint
     assert_equal(true, path     .tainted?)
@@ -749,6 +785,14 @@ class TestPathname < Test::Unit::TestCase
     assert_kind_of(Time, Pathname(__FILE__).atime)
   end
 
+  def test_birthtime
+    assert_kind_of(Time, Pathname(__FILE__).birthtime)
+  rescue NotImplementedError
+    assert_raise(NotImplementedError) do
+      File.birthtime(__FILE__)
+    end
+  end
+
   def test_ctime
     assert_kind_of(Time, Pathname(__FILE__).ctime)
   end
@@ -1222,7 +1266,7 @@ class TestPathname < Test::Unit::TestCase
       Pathname("d").mkdir
       assert(File.directory?("d"))
       Pathname("d").rmdir
-      assert(!File.exists?("d"))
+      assert(!File.exist?("d"))
     }
   end
 
@@ -1253,6 +1297,31 @@ class TestPathname < Test::Unit::TestCase
       assert_equal([Pathname("."), Pathname("a"), Pathname("b"), Pathname("d"), Pathname("d/x"), Pathname("d/y")], a)
       a = Pathname("d").find.sort
       assert_equal([Pathname("d"), Pathname("d/x"), Pathname("d/y")], a)
+
+      begin
+        File.unlink("d/y")
+        File.chmod(0600, "d")
+        a = []; Pathname(".").find(ignore_error: true) {|v| a << v }; a.sort!
+        assert_equal([Pathname("."), Pathname("a"), Pathname("b"), Pathname("d"), Pathname("d/x")], a)
+        a = []; Pathname("d").find(ignore_error: true) {|v| a << v }; a.sort!
+        assert_equal([Pathname("d"), Pathname("d/x")], a)
+
+        skip "no meaning test on Windows" if /mswin|mingw/ =~ RUBY_PLATFORM
+        a = [];
+        assert_raise_with_message(Errno::EACCES, %r{d/x}) do
+          Pathname(".").find(ignore_error: false) {|v| a << v }
+        end
+        a.sort!
+        assert_equal([Pathname("."), Pathname("a"), Pathname("b"), Pathname("d"), Pathname("d/x")], a)
+        a = [];
+        assert_raise_with_message(Errno::EACCES, %r{d/x}) do
+          Pathname("d").find(ignore_error: false) {|v| a << v }
+        end
+        a.sort!
+        assert_equal([Pathname("d"), Pathname("d/x")], a)
+      ensure
+        File.chmod(0700, "d")
+      end
     }
   end
 
@@ -1314,4 +1383,17 @@ class TestPathname < Test::Unit::TestCase
       assert_equal("foo/bar", File.join(Pathname.new("foo"), Pathname.new("bar").taint))
     }.call
   end
+
+  def test_relative_path_from_casefold
+    assert_separately([], <<-'end;') #    do
+      module File::Constants
+        remove_const :FNM_SYSCASE
+        FNM_SYSCASE = FNM_CASEFOLD
+      end
+      require 'pathname'
+      foo = Pathname.new("fo\u{f6}")
+      bar = Pathname.new("b\u{e4}r".encode("ISO-8859-1"))
+      assert_instance_of(Pathname, foo.relative_path_from(bar))
+    end;
+  end
 end
